/* gTask
 * Copyright (C) 2004 Michael Henson                                       
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <libart_lgpl/libart.h>
#include <eel/eel.h>

#include <gtask-ui/gtask-cell-renderer-basic.h>

#include "gtask-internals.h"
#include "gtask-ui-util.h"

/** @addtogroup GTaskCellRendererBasic */
/* @{ */

/** @todo implement progress pulse */

/** @todo inherit this */
#define PADDING 6

enum {
  PROP_0,
  C_PROP_PROGRESS,
  C_PROP_TITLE,
  C_PROP_STATUS_MESSAGE,
  C_PROP_ACTIVITY_STATE,
  C_PROP_TIME_LEFT,
  C_PROP_PREVIEW,
  C_PROP_PREVIEW_IS_THUMB,
  C_PROP_THUMBNAIL_FRAME
}; 

typedef struct _GTaskBasicInfo {
    PangoLayout *title_prefix;
    PangoLayout *title;
    int          title_width;
    int          title_height;

    PangoLayout *status_prefix;
    PangoLayout *status;
    int          status_width;
    int          status_height;

    PangoLayout *time_prefix;
    PangoLayout *time;
    int          time_width;
    int          time_height;

    PangoLayout *actions_layout;
    int          actions_width;
    int          actions_height;

    GdkPixbuf   *preview;
    int          preview_width;
    int          preview_height;

    int         center_x;
    int         right_x;

    int         max_title_width;
    int         text_spacer_width;
    int         text_spacer_height;
} GTaskBasicInfo;

static gpointer parent_class;

/* forward declarations */
static GTaskBasicInfo *
gtask_cell_renderer_basic_get_sizing_info( GTaskCellRendererBasic *cell,
                                           GtkWidget *widget,
                                           guint x,
                                           guint y );
static void
gtask_cell_renderer_basic_fill_text_info( GTaskCellRendererBasic *cell,
                                          GTaskBasicInfo *info,
                                          GtkWidget *widget,
                                          gint center_size );

static void
gtask_basic_info_free( GTaskBasicInfo *info );

static void
gtask_cell_renderer_basic_init( GTaskCellRendererBasic *cell );

static void
gtask_cell_renderer_basic_class_init( GTaskCellRendererBasicClass *klass );

static void
gtask_cell_renderer_basic_finalize( GObject *object );

static void
gtask_cell_renderer_basic_get_property( GObject    *object,
                                        guint       param_id,
                                        GValue     *value,
                                        GParamSpec *pspec );

static void
gtask_cell_renderer_basic_set_property( GObject      *object,
                                        guint         param_id,
                                        const GValue *value,
                                        GParamSpec   *pspec );

static void
gtask_cell_renderer_basic_get_size( GtkCellRenderer *cell,
                                    GtkWidget       *widget,
                                    GdkRectangle    *cell_area,
                                    gint            *x_offset,
                                    gint            *y_offset,
                                    gint            *width,
                                    gint            *height );

static void
gtask_cell_renderer_basic_render( GtkCellRenderer *cell,
                                  GdkWindow       *window,
                                  GtkWidget       *widget,
                                  GdkRectangle    *background_area,
                                  GdkRectangle    *cell_area,
                                  GdkRectangle    *expose_area,
                                  guint            flags );

GtkType gtask_cell_renderer_basic_get_type( ) {
	static GtkType type = 0;

	if( !type ) {
		static const GTypeInfo info = {
			sizeof( GTaskCellRendererBasicClass ),
			NULL,		/* base_init */
			NULL,		/* base_finalize */
			(GClassInitFunc) gtask_cell_renderer_basic_class_init,
			NULL,		/* class_finalize */
			NULL,		/* class_data */
			sizeof( GTaskCellRendererBasic ),
			0,          /* n_preallocs */
			(GInstanceInitFunc) gtask_cell_renderer_basic_init,
		};

		type = g_type_register_static( GTK_TYPE_CELL_RENDERER,
                                       "GTaskCellRendererBasic",
                                       &info,
                                       0);
	}

	return type;
}

static void
gtask_cell_renderer_basic_init( GTaskCellRendererBasic *cell ) {
    GTaskCellRendererBasicPrivate *private;
    gchar                         *uri;

    private = g_new0( GTaskCellRendererBasicPrivate, 1 );

    private->progress = 0.0;
    private->title = NULL;
    private->status_msg = NULL;
    private->activity_state = GTASK_ACTIVITY_STATE_INACTIVE;
    private->time_left = GTASK_TIME_LEFT_UNKNOWN;

    cell->private = private;

    /** @todo should we se our own default thumbnail frame or
     * just let the user set it -- also -- eventually we need to
     * handle frame theming like nautilus seems to do
     */

    uri = g_build_filename( DATADIR,
                            "pixmaps",
                            "default-thumbnail-frame.png",
                            NULL );

    private->thumbnail_frame = gdk_pixbuf_new_from_file( uri, NULL );

    g_free( uri );
}

static void
gtask_cell_renderer_basic_class_init( GTaskCellRendererBasicClass *klass ) {
	GObjectClass         *object_class = G_OBJECT_CLASS( klass );
	GtkCellRendererClass *cell_class = GTK_CELL_RENDERER_CLASS( klass );

	parent_class = g_type_class_peek_parent( klass );
  
	object_class->finalize = gtask_cell_renderer_basic_finalize;
  
	object_class->get_property = gtask_cell_renderer_basic_get_property;
	object_class->set_property = gtask_cell_renderer_basic_set_property;

	cell_class->get_size = gtask_cell_renderer_basic_get_size;
	cell_class->render = gtask_cell_renderer_basic_render;
  
	g_object_class_install_property( object_class,
                                     C_PROP_PROGRESS,
                                     g_param_spec_float( "progress",
                                                         "Progress",
                                                         "Value of the progress bar.",
                                                         0.0, 1.0, 0.0,
                                                         G_PARAM_READWRITE ) );
	g_object_class_install_property( object_class,
                                     C_PROP_TITLE,
                                     g_param_spec_string( "title",
                                                          "Title",
                                                          "Title of the task.",
                                                          "",
                                                          G_PARAM_READWRITE ) );

	g_object_class_install_property( object_class,
                                     C_PROP_STATUS_MESSAGE,
                                     g_param_spec_string( "status-message",
                                                          "Status Message",
                                                          "Status message of the task.",
                                                          "",
                                                          G_PARAM_READWRITE ) );

	g_object_class_install_property( object_class,
                                     C_PROP_ACTIVITY_STATE,
                                     g_param_spec_enum( "activity-state",
                                                        "Task Activity State",
                                                        "The activity state of the task.",
                                                        GTASK_ACTIVITY_STATE_TYPE,
                                                        GTASK_ACTIVITY_STATE_INACTIVE,
                                                        G_PARAM_READWRITE ) );

	g_object_class_install_property( object_class,
                                     C_PROP_PREVIEW,
                                     g_param_spec_object( "preview",
                                                          "Preview",
                                                          "A thumbnail of the resultant file or something indicative of the task",
                                                          GDK_TYPE_PIXBUF,
                                                          G_PARAM_READWRITE ) );
	g_object_class_install_property( object_class,
                                     C_PROP_PREVIEW_IS_THUMB,
                                     g_param_spec_boolean( "preview-is-thumb",
                                                          "Preview is a Thumbnail",
                                                          "Whether or not the preview is a thumbnail (ie should we draw a frame around it or not",
                                                          TRUE,
                                                          G_PARAM_READWRITE ) );
	g_object_class_install_property( object_class,
                                     C_PROP_THUMBNAIL_FRAME,
                                     g_param_spec_object( "thumbnail-frame",
                                                          "Thumbnail Frame",
                                                          "The frame to place around all generated thumbnails",
                                                          GDK_TYPE_PIXBUF,
                                                          G_PARAM_READWRITE ) );
	g_object_class_install_property( object_class,
                                     C_PROP_TIME_LEFT,
                                     g_param_spec_long( "time-left",
                                                        "Time Left",
                                                        "The remaining time in seconds until the task is expected to be finished. If the time left is unknown use GTASK_TIME_LEFT_UNKNOWN (-1)",
                                                        -1,
                                                        G_MAXLONG,
                                                        GTASK_TIME_LEFT_UNKNOWN,
                                                        G_PARAM_READWRITE ) );
}

static void
gtask_cell_renderer_basic_get_property( GObject    *object,
                                        guint       param_id,
                                        GValue     *value,
                                        GParamSpec *pspec )
{
	GTaskCellRendererBasic        *cell = GTASK_CELL_RENDERER_BASIC( object );
    GTaskCellRendererBasicPrivate *private = cell->private;

	switch( param_id ) {
		case C_PROP_PROGRESS:
			g_value_set_float( value, private->progress );
			break;
        case C_PROP_TITLE:
            g_value_set_string( value, private->title );
            break;
        case C_PROP_STATUS_MESSAGE:
            g_value_set_string( value, private->status_msg );
            break;
        case C_PROP_ACTIVITY_STATE:
            g_value_set_int( value, private->activity_state );
            break;
        /** @todo how will memory management work with this? */
        case C_PROP_PREVIEW:
            g_value_set_object( value, private->preview );
            break;
        case C_PROP_PREVIEW_IS_THUMB:
            g_value_set_boolean( value, private->preview_is_thumb );
            break;
        case C_PROP_TIME_LEFT:
            g_value_set_long( value, private->time_left );
            break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID( object, param_id, pspec );
	}
}

static void
gtask_cell_renderer_basic_set_property( GObject       *object,
                                        guint         param_id,
                                        const GValue *value,
                                        GParamSpec   *pspec )
{
	GTaskCellRendererBasic        *cell = GTASK_CELL_RENDERER_BASIC( object );
    GTaskCellRendererBasicPrivate *private = cell->private;

	switch( param_id ) {
		case C_PROP_PROGRESS:
			private->progress = g_value_get_float( value );
	        g_object_notify( object, "progress" );
			break;
        case C_PROP_TITLE:
            if( private->title )
                g_free( private->title );

            private->title = g_strdup( g_value_get_string( value ) );
            g_object_notify( object, "title" );
            break;
        case C_PROP_STATUS_MESSAGE:
            if( private->status_msg )
                g_free( private->status_msg );

            private->status_msg = g_strdup( g_value_get_string( value ) );
            g_object_notify( object, "status-message" );
            break;
        case C_PROP_ACTIVITY_STATE:
            private->activity_state = g_value_get_enum( value );
            g_object_notify( object, "activity-state" );
            break;
        case C_PROP_PREVIEW:
            private->preview = g_value_get_object( value );
            g_object_notify( object, "preview" );
            break;
        case C_PROP_PREVIEW_IS_THUMB:
            private->preview_is_thumb = g_value_get_boolean( value );
            g_object_notify( object, "preview-is-thumb" );
            break;
        case C_PROP_TIME_LEFT:
            private->time_left = g_value_get_long( value );
            g_object_notify( object, "time-left" );
            break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID( object, param_id, pspec );
	}
}

#define ELLIPSIS "..."

/* the following is a variation on the ellipsizing-label from eel,
 * I basically just added support for marking up portions of the string */

/* the following modified from eel-pango-extensions.c */
static int
measure_string_width (const char  *string, PangoLayout *layout) {
    int width;

    pango_layout_set_markup (layout, string, -1);
    pango_layout_get_pixel_size (layout, &width, NULL);

    return width;
}

static void
compute_character_widths( const char    *format_string,
                          const char    *string,
                          PangoLayout   *layout,
                          int           *char_len_return,
                          int          **widths_return,
                          int          **cuts_return )
{
    int *widths;
    int *offsets;
    int *cuts;
    int char_len;
    int byte_len;
    const char *p;
    int i;
    PangoLayoutIter *iter;
    PangoLogAttr *attrs;

#define BEGINS_UTF8_CHAR(x) (((x) & 0xc0) != 0x80)

    char_len = g_utf8_strlen( string, -1 );
    byte_len = strlen( string );

    widths = g_new (int, char_len);
    offsets = g_new (int, byte_len);

    /* Create a translation table from byte index to char offset */
    p = string;
    i = 0;
    while (*p) {
        int byte_index = p - string;

        if (BEGINS_UTF8_CHAR (*p)) {
            offsets[byte_index] = i;
            ++i;
        } else {
            offsets[byte_index] = G_MAXINT; /* segv if we try to use this */
        }

        ++p;
    }

    /* Now fill in the widths array */
    pango_layout_set_markup (layout, format_string, -1);
   iter = pango_layout_get_iter (layout);

    do {
        PangoRectangle extents;
        int byte_index;

        byte_index = pango_layout_iter_get_index (iter);

        if (byte_index < byte_len) {
            pango_layout_iter_get_char_extents (iter, &extents);

            g_assert (BEGINS_UTF8_CHAR (string[byte_index]));
            g_assert (offsets[byte_index] < char_len);

            widths[offsets[byte_index]] = PANGO_PIXELS (extents.width);
        }

    } while (pango_layout_iter_next_char (iter));

    pango_layout_iter_free (iter);

    g_free (offsets);

    *widths_return = widths;

    /* Now compute character offsets that are legitimate places to
     * chop the string
     */
    attrs = g_new (PangoLogAttr, char_len + 1);

    pango_get_log_attrs (string, byte_len, -1,
                 pango_context_get_language (
                     pango_layout_get_context (layout)),
                 attrs,
                 char_len + 1);

    cuts = g_new (int, char_len);
    i = 0;
    while (i < char_len) {
        cuts[i] = attrs[i].is_cursor_position;

        ++i;
    }

    g_free (attrs);
    *cuts_return = cuts;

    *char_len_return = char_len;
}

static char *
eel_string_ellipsize_end (const char *format_string,
                          const char *string,
                          PangoLayout *layout,
                          int width)
{
    int resulting_width;
    int *cuts;
    int *widths;
    int char_len;
    const char *p;
    int truncate_offset;
    char *result;

    /* See explanatory comments in ellipsize_start */

    if ( !string || *string == '\0')
        return g_strdup ("");

    resulting_width = measure_string_width (format_string, layout);

    if (resulting_width <= width) {
        return g_strdup (string);
    }

    width -= measure_string_width (ELLIPSIS, layout);

    if (width < 0) {
        return g_strdup ("");
    }

    compute_character_widths (format_string, string, layout, &char_len, &widths, &cuts);

        for (truncate_offset = char_len - 1; truncate_offset > 0; truncate_offset--) {
            resulting_width -= widths[truncate_offset];
            if (resulting_width <= width &&
            cuts[truncate_offset]) {
            break;
            }
        }

    g_free (cuts);
    g_free (widths);

    p = g_utf8_offset_to_pointer (string, truncate_offset);

    result = g_malloc ((p - string) + strlen (ELLIPSIS) + 1);
    memcpy (result, string, (p - string));
    strcpy (result + (p - string), ELLIPSIS);

    return result;
}
/* end of functions borrowed and modified from eel-pango-extensions.c */

/* have optional prefix layout, always request that plus 10 blank
 * spaces of space
 *
 * then run string through a slightly modified eel-pango foo
 * to chop off what needs to go
 */

static GTaskBasicInfo *
gtask_cell_renderer_basic_get_sizing_info( GTaskCellRendererBasic *cell,
                                           GtkWidget *widget,
                                           guint x,
                                           guint y )
{
    GTaskCellRendererBasicPrivate *private = cell->private;
    GTaskBasicInfo                *info    = g_new0( GTaskBasicInfo, 1 );
    PangoLayout                   *spacer_layout;

    info->title_width = 0;
    info->title_height = 0;

    spacer_layout = gtk_widget_create_pango_layout( widget, NULL );

    pango_layout_set_markup( spacer_layout, 
                             "<span size=\"medium\">          </span>",
                             -1 );
    pango_layout_get_pixel_size( spacer_layout,
                                 &(info->text_spacer_width),
                                 &(info->text_spacer_height) );

    info->status_prefix = gtk_widget_create_pango_layout( widget, NULL );
    pango_layout_set_markup( info->status_prefix,
                             "<span size=\"small\">    <b>Info:</b> </span>",
                             -1 );

    pango_layout_get_pixel_size( info->status_prefix,
                                 &(info->status_width),
                                 &(info->status_height) );

    info->time_prefix = gtk_widget_create_pango_layout( widget, NULL );
    pango_layout_set_markup( info->time_prefix,
                            "<span size=\"small\">    <b>Time Left:</b> </span>",
                            -1 );

    pango_layout_get_pixel_size( info->time_prefix,
                                 &(info->time_width),
                                 &(info->time_height) );

    info->actions_layout = gtk_widget_create_pango_layout( widget, NULL );
    pango_layout_set_markup( info->actions_layout,
                             "<span size=\"small\"><b>Actions:</b></span>",
                             -1 );

    pango_layout_get_pixel_size( info->actions_layout,
                                 &(info->actions_width),
                                 &(info->actions_height) );

    info->preview_height = info->text_spacer_height + info->status_height + info->time_height;
    info->preview_width = info->preview_height;

    if( private->preview )
        info->preview = private->preview;

    info->center_x = x + PADDING + info->preview_width + 12;
    info->right_x = info->center_x + MAX( info->title_width,
                                          MAX( info->status_width, 
                                               info->time_width ) ) + 12;


    info->max_title_width = MAX( info->status_width, info->time_width );

    return info;
}

void
gtask_cell_renderer_basic_fill_text_info( GTaskCellRendererBasic *cell,
                                          GTaskBasicInfo *info,
                                          GtkWidget *widget,
                                          gint center_width )
{
    GTaskCellRendererBasicPrivate *private = cell->private;
    PangoLayout                   *layout;
    GString                       *markup  = g_string_new( NULL );
    GString                       *tmp = g_string_new( NULL );
    char                          *text;

    /* title */
    info->title = layout = gtk_widget_create_pango_layout( widget, NULL );

    g_string_printf( markup, "<span size=\"medium\"><b>%s</b></span>",
                     private->title );

    text = eel_string_ellipsize_end( markup->str, private->title, layout, center_width );

    g_string_printf( markup, "<span size=\"medium\"><b>%s</b></span>",
                     text );

    pango_layout_set_markup( layout, markup->str, -1 );

    g_free( text );

    /* info */
    info->status = layout = gtk_widget_create_pango_layout( widget, NULL );

    /** @todo check if there is an error message set and display that in
     *        some error color -- perhaps bold red? -- and Info: should be
     *        replaced with Error: */
    g_string_printf( markup, "<span size=\"small\">%s</span>",
                     private->status_msg );

    text = eel_string_ellipsize_end( markup->str, private->status_msg, layout, center_width - info->status_width );

    g_string_printf( markup, "<span size=\"small\">%s</span>",
                     text );

    pango_layout_set_markup( layout, markup->str, -1 );

    g_free( text );

    /* time left */
    info->time = layout = gtk_widget_create_pango_layout( widget, NULL );

    if( private->activity_state == GTASK_ACTIVITY_STATE_COMPLETED ) {
        g_string_assign( markup, "<span size=\"small\">None</span>" );
        g_string_assign( tmp, "None" );
    } else if( private->time_left < 0 ) {
        g_string_assign( markup, "<span size=\"small\">Unknown</span>" );
        g_string_assign( tmp, "Unknown" );
    } else {
        g_string_printf( markup, "<span size=\"small\">%ld seconds</span>", private->time_left );
        g_string_printf( tmp, "%ld seconds", private->time_left );
    }

    text = eel_string_ellipsize_end( markup->str, tmp->str, layout, center_width - info->time_width );

    g_string_printf( markup, "<span size=\"small\">%s</span>", tmp->str );

    pango_layout_set_markup( layout, markup->str, -1 );

    g_free( text );

    g_string_free( markup, TRUE );
    g_string_free( tmp, TRUE );
}

static void
gtask_basic_info_free( GTaskBasicInfo *info ) {
    if( info->title )
        g_object_unref( info->title );

    if( info->status_prefix )
        g_object_unref( info->status_prefix );

    if( info->status )
        g_object_unref( info->status );

    if( info->time_prefix )
        g_object_unref( info->time_prefix );

    if( info->time )
        g_object_unref( info->time );

    if( info->actions_layout )
        g_object_unref( info->actions_layout );

//    gdk_pixbuf_unref( info->preview );

    g_free( info );
}

static void
gtask_cell_renderer_basic_get_size( GtkCellRenderer *cell,
                                    GtkWidget       *widget,
                                    GdkRectangle    *cell_area,
                                    gint            *x_offset,
                                    gint            *y_offset,
                                    gint            *width,
                                    gint            *height)
{
    GTaskCellRendererBasic *cell_basic = GTASK_CELL_RENDERER_BASIC( cell );
    GTaskBasicInfo *info = gtask_cell_renderer_basic_get_sizing_info( cell_basic, widget, 0, 0 );

    g_debug( "IN: gtask_cell_renderer_basic_get_size" );

	if( width ) {
        *width = PADDING + info->preview_width + 8 +
                 info->max_title_width + info->text_spacer_width + 8 +
                 MAX( 100,
                      MAX( info->actions_width,
                          4 * 8 + 3 * info->time_height ) ) +
                 PADDING;
    }
	if( height ) {
		*height = PADDING +
                  MAX( info->preview_height,
                       info->text_spacer_height + 4 + info->status_height +
                       info->time_height ) +
                  PADDING;
    }

    gtask_basic_info_free( info );

    g_debug( "LEAVING: gtask_cell_renderer_basic_get_size" );
}

GtkCellRenderer*
gtask_cell_renderer_basic_new( void ) {
	return GTK_CELL_RENDERER( g_object_new( GTASK_CELL_RENDERER_BASIC_TYPE, NULL ) );
}

static void
gtask_cell_renderer_basic_render( GtkCellRenderer *cell,
                                  GdkWindow       *window,
                                  GtkWidget       *widget,
                                  GdkRectangle    *background_area,
                                  GdkRectangle    *cell_area,
                                  GdkRectangle    *expose_area,
                                  guint            flags)
{
	GTaskCellRendererBasic *cell_basic = GTASK_CELL_RENDERER_BASIC( cell );
	GTaskCellRendererBasicPrivate *private = cell_basic->private;
    GTaskBasicInfo         *info = gtask_cell_renderer_basic_get_sizing_info( cell_basic, widget, cell_area->x, cell_area->y );

    
    guint x = cell_area->x;
    guint y = cell_area->y;

	GdkGC *gc;

    GtkStateType   gtk_state;

    GTaskActivityState activity_state = cell_basic->private->activity_state;

    g_debug( "IN: gtask_cell_renderer_basic_render" );

    /** @todo the size of the progress bar may grow for larger fonts */
    gtask_cell_renderer_basic_fill_text_info( cell_basic,
                                              info,
                                              widget,
                                              cell_area->width - info->preview_width - 100 - 16 - 2 * PADDING );

    if( activity_state == GTASK_ACTIVITY_STATE_INACTIVE )
        gtk_state = GTK_STATE_INSENSITIVE;
    else {
        switch( flags ) {
            case GTK_CELL_RENDERER_SELECTED:
                gtk_state = GTK_STATE_SELECTED;
                break;
            default:
                gtk_state = GTK_STATE_NORMAL;
                break;
        }
    }

    gc = gdk_gc_new( window );

    if( info->preview ) {
        /** @todo cell renderers really need to have some sort of cache
         *        data associated with them so when know when we really
         *        need to redo this or not
         */

        guint p_x, p_y;
        guint width, height;
        guint max_size = info->text_spacer_height + info->status_height + info->time_height;

        GdkPixbuf *scaled = eel_gdk_pixbuf_scale_down_to_fit( info->preview,
                                                              max_size,
                                                              max_size );
        GdkPixbuf *final;

        if( private->thumbnail_frame && private->preview_is_thumb ) {
            final = eel_embed_image_in_frame( scaled,
                                              private->thumbnail_frame,
                                              3, 3, 6, 6 );

            g_object_unref( (GObject *) scaled );
        } else {
            final = scaled;
        }

        width = gdk_pixbuf_get_width( final );
        height = gdk_pixbuf_get_height( final );

        /* place icons in the middle of the "box" */
        p_x = x + ( max_size - width ) / 2;
        p_y = y +( max_size - height ) / 2;

        if( private->activity_state == GTASK_ACTIVITY_STATE_INACTIVE ) {
            /* grey out the resulting image */
            gdk_pixbuf_saturate_and_pixelate( final, final, 1.0, TRUE );
        }

        gdk_pixbuf_render_to_drawable( final, window, gc, 0, 0,
                                       p_x + PADDING, p_y + PADDING, -1, -1,
                                       GDK_RGB_DITHER_NONE, -1, -1 );

        g_object_unref( (GObject *) final );


        /** @todo -- do we need to do placement recalcs now? */
    }

	gtk_paint_layout( widget->style, window,
                      gtk_state,
                      FALSE,
                      cell_area,
                      widget,
                      "title",
                      info->center_x,
                      y + PADDING + 2,
                      info->title );

	gtk_paint_layout( widget->style, window,
                        gtk_state,
                        FALSE,
                        cell_area,
                        widget,
                        "status-message-label",
                        info->center_x,
                        y + PADDING + info->text_spacer_height + 4,
                        info->status_prefix );

	gtk_paint_layout( widget->style, window,
                        gtk_state,
                        FALSE,
                        cell_area,
                        widget,
                        "status-message",
                        info->center_x + info->status_width,
                        y + PADDING + info->text_spacer_height + 4,
                        info->status );

	gtk_paint_layout( widget->style, window,
                        gtk_state,
                        FALSE,
                        cell_area,
                        widget,
                        "time-remaining-label",
                        info->center_x,
                        y + PADDING + info->text_spacer_height + 4 + info->status_height,
                        info->time_prefix );

	gtk_paint_layout( widget->style, window,
                        gtk_state,
                        FALSE,
                        cell_area,
                        widget,
                        "time-remaining",
                        info->center_x + info->time_width,
                        y + PADDING + info->text_spacer_height + 4 + info->status_height,
                        info->time );

    /** @todo properly place the progress bar based on it's size
     *        right now it's just set to 100
     */
    gtask_ui_draw_progressbar( widget, window, cell_area,
                               cell_basic->private->progress, activity_state,
                               cell_area->x + cell_area->width - 100 - PADDING, y + PADDING,
                               100, info->text_spacer_height + 4 );

	gtk_paint_layout( widget->style, window,
                        gtk_state,
                        FALSE,
                        cell_area,
                        widget,
                        "actions",
                        cell_area->x + cell_area->width - 100 - PADDING,
                        y + PADDING + info->text_spacer_height + 4,
                        info->actions_layout );

    gtask_basic_info_free( info );

    g_debug( "LEAVING: gtask_cell_renderer_basic_render" );
}

static void
gtask_cell_renderer_basic_finalize( GObject *object ) {
	GTaskCellRendererBasic *cell = GTASK_CELL_RENDERER_BASIC( object );

	g_free( cell->private );
  
	if( G_OBJECT_CLASS (parent_class)->finalize )
    	G_OBJECT_CLASS (parent_class)->finalize( object );
}

/* @} */
