/* gTask
 * Copyright (C) 2004 Michael Henson
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gtask-ui/gtask-ui-manager.h>

#include <gtask/gtask-task-packet.h>

#include "gtask-internals.h"

/** @addtogroup GTaskUIManager */

/* @{ */

#define WANT_WINDOWS_KEY "/apps/gtask/enable_per_program_window"

static gpointer parent_class;

static void
gtask_ui_manager_class_init( GTaskUIManagerClass *klass );

static void
gtask_ui_manager_init( GTaskUIManager *ui_manager, gpointer g_class );

static void
gtask_ui_manager_dispose( GObject *obj );

static void
gtask_ui_manager_finalize( GObject *obj );

static void
gtask_ui_manager_task_window_destroyed( GtkObject *obj, gpointer data );

static void
gtask_ui_manager_private_pref_change( GConfClient *client,
                                      guint conn_id,
                                      GConfEntry *entry,
                                      gpointer data);

GType gtask_ui_manager_get_type( void ) {
    static GType type = 0;

    if( type == 0 ) {
        static const GTypeInfo info = {
            sizeof( GTaskUIManagerClass ),
            NULL,   /* base init */
            NULL,   /* base_finalize */
            (GClassInitFunc) gtask_ui_manager_class_init,   /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            sizeof( GTaskUIManager ),
            0,      /* n_preallocs */
            (GInstanceInitFunc) gtask_ui_manager_init
        };

        type = g_type_register_static( G_TYPE_OBJECT,
                                       "GTaskUIManager",
                                       &info,
                                       0 );
    }

    return type;
}

static void
gtask_ui_manager_class_init( GTaskUIManagerClass *klass ) {
    GObjectClass *gobject_class = G_OBJECT_CLASS( klass );

    parent_class = g_type_class_peek_parent( klass );

    gobject_class->dispose = gtask_ui_manager_dispose;
    gobject_class->finalize = gtask_ui_manager_finalize;
}

static void
gtask_ui_manager_init( GTaskUIManager *ui_manager, gpointer g_class ) {
    GTaskUIManagerPrivate *private;

    private = g_new0( GTaskUIManagerPrivate, 1 );

    private->gconf_client = gconf_client_get_default( );
    private->task_window = NULL;
    private->display_task_window = TRUE;

    /*
    private->display_task_window = gconf_client_get_bool( private->gconf_client,
                                                          WANT_WINDOWS_KEY,
                                                          NULL );

    private->gconf_notify_id = gconf_client_notify_add( private->gconf_client,
                                           WANT_WINDOWS_KEY,
                                           gtask_ui_manager_private_pref_change,
                                           ui_manager,
                                           NULL,
                                           NULL );
    */

    ui_manager->private = private;
}

static void
gtask_ui_manager_dispose( GObject *obj ) {
    /** @todo implement me */
}

static void
gtask_ui_manager_finalize( GObject *obj ) {
    /** @todo implement me */
}

/**
 * @brief creates a new GTaskUIManager
 *
 * @author Michael Henson
 *
 * @return the new GTaskUIManager
 */
GTaskUIManager *
gtask_ui_manager_new( ) {
    return g_object_new( GTASK_UI_MANAGER_TYPE, NULL );
}

gboolean
gtask_ui_manager_private_task_event( GTaskEventDispatcher *dispatcher,
                                     GObject *i_packet,
                                     gpointer data )
{
    GTaskUIManagerPrivate *private;
    GTaskTaskPacket       *packet;
    GTaskGenericTask      *task;

    g_debug( "IN: gtask_ui_manager_private_task_event" );

    g_return_val_if_fail( GTASK_IS_UI_MANAGER( data ), FALSE );

    if( !GTASK_IS_TASK_PACKET( i_packet ) ) {
        g_debug( "hmmm . . we were not passed a task packet . . ignore" );
        return TRUE;
    }

    private = GTASK_UI_MANAGER( data )->private;
    packet = GTASK_TASK_PACKET( i_packet );

    if( !private->display_task_window ) {
        g_debug( "no task window should be displayed" );
        return TRUE;
    }

    g_debug( "displaying task window" );

    if( !private->task_window ) {
        private->task_window = gtask_window_new_with_name( NULL, NULL );

        g_signal_connect( private->task_window,
                          "destroy",
                          (GCallback) gtask_ui_manager_task_window_destroyed,
                          data );

        gtk_widget_show_all( GTK_WIDGET( private->task_window ) );
    }

    task = gtask_task_packet_get_task( packet );

//    gtask_generic_task_dump( task );

    gtask_window_add_task( private->task_window, task );

    g_debug( "LEAVING: gtask_ui_manager_private_task_event" );

    return TRUE;
}

static void
gtask_ui_manager_task_window_destroyed( GtkObject *obj, gpointer data ) {
    GTaskUIManagerPrivate *private;

    g_return_if_fail( GTASK_IS_UI_MANAGER( data ) );

    g_debug( "IN: gtask_ui_manager_task_window_destroyed" );

    private = GTASK_UI_MANAGER( data )->private;

    if( GTK_OBJECT( obj ) == GTK_OBJECT( private->task_window ) ) {
        private->task_window = NULL;
    }

    g_debug( "LEAVING: gtask_ui_manager_task_window_destroyed" );
}

static void
gtask_ui_manager_private_pref_change( GConfClient *client,
                                      guint conn_id,
                                      GConfEntry *entry,
                                      gpointer data)
{
    GTaskUIManager *ui_manager;
    const char     *key = gconf_entry_get_key( entry );
    GConfValue     *val;

    g_debug( "IN: gtask_ui_manager_private_pref_change" );

    g_warning( "GConf pref change on key: %s", key );

    g_return_if_fail( GTASK_IS_UI_MANAGER( data ) );

    if( strcmp( key, WANT_WINDOWS_KEY ) != 0 )
        return;

    ui_manager = GTASK_UI_MANAGER( data );

    val = gconf_entry_get_value( entry );

    ui_manager->private->display_task_window = gconf_value_get_bool( val );

    g_debug( "LEAVING: gtask_ui_manager_private_pref_change" );
}

/* @} */
