/* gTask
 * Copyright (C) 2004 Michael Henson
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <string.h>

#include <gdk/gdk.h>

#include <gtk/gtk.h>

#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkcellrenderertext.h>

#include <gnome.h>

#include <eel/eel.h>

#include <gtask-ui/gtask-view.h>

#include <gtask-ui/gtask-cell-renderer-basic.h>
#include <gtask-ui/gtask-cell-renderer-progress.h>

#include <gtask-ui/gtask-marshalers.h>

#include "gtask-internals.h"
#include "gtask-ui-util.h"

/** @addtogroup GTaskView */

/* @{ */

enum {
    COL_TASK,

    COL_TITLE,
    COL_CATEGORY,

    COL_STATUS_MESSAGE,
    COL_ERROR_MESSAGE,

    COL_PROGRESS,
    COL_TIME_LEFT,
    COL_ACTIVITY_STATE,

    COL_OLD_THUMB_URI,
    COL_THUMBNAIL,
    COL_IS_THUMBNAIL,

    N_TASK_COLUMNS
};

enum {
    PROP_0,
    PROP_VIEW_MODE
};

/** @todo replace this */
typedef struct _GTaskTaskIter {
    GTaskGenericTask *task;
    GtkTreeIter       iter;
} GTaskTaskIter;

struct GTaskGetTaskFoo {
    guint      num;
    GTaskView *view;
    GList     *tasks;
    gboolean   with_active;
    gboolean   with_inactive;
    gboolean   with_completed;
};

static void
gtask_task_iter_free( GTaskTaskIter *iter ) {
    g_object_unref( iter->task );
    g_free( iter );
}

static gpointer parent_class;

static void
gtask_view_init( GTaskView *view, gpointer g_class );

static void
gtask_view_class_init( GTaskViewClass *klass );

static void
gtask_view_destroy( GtkObject *obj );

static void
gtask_view_finalize( GObject *obj );

static void
gtask_view_create_basic_mode( GTaskView *view );

static void
gtask_view_create_list_mode( GTaskView *view );

static void
gtask_view_create_task_model( GTaskView *view );

static void
gtask_view_update_task( GTaskView *view, GTaskTaskIter *iter );

static gboolean
gtask_view_button_press( GtkWidget *list,
                         GdkEventButton *event,
                         gpointer data );

static void
gtask_view_open_file( GObject *obj, gpointer *data );

static void
gtask_view_open_dir( GObject *obj, gpointer *data );

/*
static void
gtask_view_open_file_properties( GObject *obj, gpointer *data );
*/

static void
gtask_view_set_property( GObject *object,
                         guint param_id,
                         const GValue *value,
                         GParamSpec *pspec );

static void
gtask_view_get_property( GObject *object,
                         guint param_id,
                         GValue *value,
                         GParamSpec *pspec );

GType gtask_view_mode_get_type( ) {
    static GType type = 0;

    if( type == 0 ) {
        static const GEnumValue values[] = {
            { GTASK_VIEW_MODE_BASIC, "GTASK_VIEW_MODE_BASIC", "basic view" },
            { GTASK_VIEW_MODE_LIST, "GTASK_VIEW_MODE_LIST", "list view" },
            { 0, NULL, NULL }
        };

        type = g_enum_register_static( "GTaskViewMode", values );
    }

    return type;
}

GType gtask_view_get_type( ) {
    static GType type = 0;

    if( type == 0 ) {
        static const GTypeInfo info = {
            sizeof( GTaskViewClass ),
            NULL,   /* base init */
            NULL,   /* base_finalize */
            (GClassInitFunc) gtask_view_class_init,   /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            sizeof( GTaskView ),
            0,      /* n_preallocs */
            (GInstanceInitFunc) gtask_view_init
            };
    
            type = g_type_register_static( GTK_TYPE_NOTEBOOK,
                                           "GTaskView",
                                           &info,
                                           0 );
        }

    return type;
}

static void
gtask_view_init( GTaskView *view, gpointer g_class ) {
    GTaskViewPrivate *private;

    private = view->private = g_new0( GTaskViewPrivate, 1 );

    private->filter_string = g_string_new( NULL );
    private->filter_seed = g_ptr_array_new( );
    private->filter_columns = g_ptr_array_new( );

    /* the columns to filter by default, in the future this may perhaps
     * be configurable -- who's to say?
     */
    g_ptr_array_add( private->filter_columns,
                     GINT_TO_POINTER( COL_TITLE ) );

    g_ptr_array_add( private->filter_columns,
                     GINT_TO_POINTER( COL_STATUS_MESSAGE ) );

    g_ptr_array_add( private->filter_columns,
                     GINT_TO_POINTER( COL_ERROR_MESSAGE ) );

    g_ptr_array_add( private->filter_columns,
                     GINT_TO_POINTER( COL_CATEGORY ) );

    private->task_hash = g_hash_table_new_full( g_str_hash, g_str_equal,
                                                g_free,
                                                (GDestroyNotify) gtask_task_iter_free );

    gtk_notebook_set_show_tabs( (GtkNotebook *) view, FALSE );
    gtk_notebook_set_show_border( (GtkNotebook *) view, TRUE );

    gtask_view_create_task_model( view );
    gtask_view_create_basic_mode( view );
    gtask_view_create_list_mode( view );

    private->view_mode = GTASK_VIEW_MODE_BASIC;

    gtk_tree_view_set_model( private->task_view_basic,
                             (GtkTreeModel *) private->task_model_filter );

    gtk_notebook_set_current_page( (GtkNotebook *) view, 0 );

    gtk_tree_view_set_model( private->task_view_list, NULL );

    private->thumbnailer = gnome_thumbnail_factory_new( GNOME_THUMBNAIL_SIZE_NORMAL );
}

static void
gtask_view_class_init( GTaskViewClass *klass ) {
    GObjectClass *gobject_class = G_OBJECT_CLASS( klass );
    GtkObjectClass *gtk_object_class = GTK_OBJECT_CLASS( klass );
    GType           param_types[1];

    parent_class = g_type_class_peek_parent( klass );

    gobject_class->finalize = gtask_view_finalize;
    gobject_class->set_property = gtask_view_set_property;
    gobject_class->get_property = gtask_view_get_property;

    gtk_object_class->destroy = gtask_view_destroy;

    g_object_class_install_property( gobject_class,
                                     PROP_VIEW_MODE,
                                     g_param_spec_enum( "view-mode",
                                                        "View Mode",
                                                        "The currently selected view mode for the GTaskView",
                                                        GTASK_VIEW_MODE_TYPE,
                                                        GTASK_VIEW_MODE_BASIC,
                                                        G_PARAM_READWRITE ) );

    param_types[0] = G_TYPE_POINTER;

    klass->tasks_added_signal_id = g_signal_newv( "tasks-added",
                                                  G_TYPE_FROM_CLASS( klass ),
                                                  G_SIGNAL_RUN_LAST |
                                                    G_SIGNAL_NO_RECURSE |
                                                    G_SIGNAL_NO_HOOKS,
                                                  NULL,
                                                  NULL,
                                                  NULL,
                                                  gtask_marshal_VOID__POINTER,
                                                  G_TYPE_NONE,
                                                  1,
                                                  param_types );

    klass->tasks_updated_signal_id = g_signal_newv( "tasks-updated",
                                                  G_TYPE_FROM_CLASS( klass ),
                                                  G_SIGNAL_RUN_LAST |
                                                    G_SIGNAL_NO_RECURSE |
                                                    G_SIGNAL_NO_HOOKS,
                                                  NULL,
                                                  NULL,
                                                  NULL,
                                                  gtask_marshal_VOID__POINTER,
                                                  G_TYPE_NONE,
                                                  1,
                                                  param_types );

    klass->tasks_removed_signal_id = g_signal_newv( "tasks-removed",
                                                  G_TYPE_FROM_CLASS( klass ),
                                                  G_SIGNAL_RUN_LAST |
                                                    G_SIGNAL_NO_RECURSE |
                                                    G_SIGNAL_NO_HOOKS,
                                                  NULL,
                                                  NULL,
                                                  NULL,
                                                  gtask_marshal_VOID__POINTER,
                                                  G_TYPE_NONE,
                                                  1,
                                                  param_types );
}

static void
gtask_view_destroy( GtkObject *obj ) {
    //GTaskView *view = GTASK_VIEW( obj );

    if( GTK_OBJECT_CLASS( parent_class )->destroy )
        GTK_OBJECT_CLASS( parent_class )->destroy( obj );
}

static void
gtask_view_finalize( GObject *obj ) {
    GTaskView *view = GTASK_VIEW( obj );

    g_free( view->private );

    g_string_free( view->private->filter_string, TRUE );

    g_ptr_array_free( view->private->filter_seed, TRUE );

    if( G_OBJECT_CLASS( parent_class )->finalize )
        G_OBJECT_CLASS( parent_class )->finalize( obj );
}

static gboolean
gtask_view_filter_func( GtkTreeModel *model, GtkTreeIter *iter, gpointer data )
{
    GTaskView        *view = GTASK_VIEW( data );
    GTaskViewPrivate *private = view->private;

    gpointer         *columns = private->filter_columns->pdata;
    GPtrArray        *filter_seed = private->filter_seed;
    guint             n_columns = private->filter_columns->len;
    guint             i;
    guint             j;
    char             *low_str;
    char             *test_str;

    GString          *tot_str;

    g_debug( "IN: gtask_view_filter_func" );

    /* end right now if there is no filter string */
    if( filter_seed->len == 0 ) {
        g_debug( "LEAVING: gtask_view_filter_func - filter_seed empty" );
        return TRUE;
    }

    /* also end if no filter columns have been set */
    if( n_columns < 1 ) {
        g_debug( "LEAVING: gtask_view_filter_func - no filter columns set" );
        return TRUE;
    }

    tot_str = g_string_new( NULL );

    /* right now we are just placing everything in one larger string
     * and matching substrings on that. In cases where the model might
     * old a wheee bit of text this might not be that bright of an idea
     * @todo determine a possible better way of doing this 
     */
    for( i = 0; i < n_columns; i++ ) {
        const char *val;
        GValue      value = { 0, };

        gtk_tree_model_get_value( GTK_TREE_MODEL( private->task_model ), iter,
                                  GPOINTER_TO_INT( columns[i] ), &value );

        val = g_value_get_string( &value );

        if( !val ) {
            g_value_unset( &value );

            continue;
        }

        g_string_append( tot_str, val );

        g_value_unset( &value );
    }

    low_str = g_utf8_strdown( tot_str->str, -1 );
    test_str = g_utf8_normalize( low_str, -1, G_NORMALIZE_ALL );

    g_string_free( tot_str, TRUE );
    g_free( low_str );

    for( j = 0; j < filter_seed->len; j++ ) {
        char *seed_tmp = filter_seed->pdata[j];

        if( g_strrstr( test_str, seed_tmp ) == NULL ) {
            g_free( test_str );

            g_debug( "LEAVING: gtask_view_filter_func - row will not be shown" );
            return FALSE;
        }
    }

    g_free( test_str );
    g_debug( "LEAVING: gtask_view_filter_func - row will be shown" );

    return TRUE;
}

static void
gtask_view_create_task_model( GTaskView *view ) {
    GTaskViewPrivate   *private= view->private;
    GtkTreeStore       *task_model;
    GtkTreeModelFilter *task_model_filter;
                
    g_assert( private->task_model == NULL );
    g_assert( private->task_model_filter == NULL );
                        
    task_model = gtk_tree_store_new( N_TASK_COLUMNS,
                                     GTASK_GENERIC_TASK_TYPE,

                                     G_TYPE_STRING,
                                     G_TYPE_STRING,

                                     G_TYPE_STRING,
                                     G_TYPE_STRING,

                                     G_TYPE_FLOAT,
                                     G_TYPE_LONG,
                                     GTASK_ACTIVITY_STATE_TYPE,
                                     
                                     G_TYPE_STRING,
                                     GDK_TYPE_PIXBUF,
                                     G_TYPE_BOOLEAN );

    task_model_filter = GTK_TREE_MODEL_FILTER( gtk_tree_model_filter_new( GTK_TREE_MODEL( task_model ), NULL ) );

    gtk_tree_model_filter_set_visible_func( task_model_filter,
                                            gtask_view_filter_func,
                                            view,
                                            NULL );

    private->task_model = task_model;
    private->task_model_filter = task_model_filter;
}

static void
gtask_view_create_basic_mode( GTaskView *view ) {
    GTaskViewPrivate   *private = view->private;
    GtkTreeViewColumn  *info_column;
    GtkCellRenderer    *rend;
    GtkTreeView        *task_view;
    GtkWidget          *task_scroll;

    g_assert( private->task_view_basic == NULL );

    task_scroll = gtk_scrolled_window_new( NULL, NULL );
    gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW( task_scroll ),
                                    GTK_POLICY_NEVER,
                                    GTK_POLICY_AUTOMATIC );

    task_view = GTK_TREE_VIEW( gtk_tree_view_new( ) );

    gtk_container_add( GTK_CONTAINER( task_scroll ), GTK_WIDGET( task_view ) );

    gtk_notebook_append_page( (GtkNotebook *) view,
                              GTK_WIDGET( task_scroll ),
                              NULL );

    gtk_tree_view_set_headers_visible( task_view, FALSE );
    gtk_tree_view_set_rules_hint( task_view, TRUE );
    gtk_tree_view_set_reorderable( task_view, TRUE );

    info_column = gtk_tree_view_column_new();

    rend = GTK_CELL_RENDERER( gtask_cell_renderer_basic_new( ) );

    gtk_tree_view_column_pack_start( info_column, rend, TRUE );
    gtk_tree_view_column_set_attributes( info_column, rend,
                                         "progress", COL_PROGRESS,
                                         "title", COL_TITLE,
                                         "status-message", COL_STATUS_MESSAGE,
                                         "activity-state", COL_ACTIVITY_STATE,
                                         "preview", COL_THUMBNAIL,
                                         "preview-is-thumb", COL_IS_THUMBNAIL,
                                         "time-left", COL_TIME_LEFT,
                                         NULL );

    gtk_tree_view_append_column( GTK_TREE_VIEW( task_view ), info_column );

    g_signal_connect( G_OBJECT( task_view ),
                      "button-press-event",
                      G_CALLBACK( gtask_view_button_press ),
                      view );

    private->task_view_basic = task_view;
}

static void
gtask_view_create_list_mode( GTaskView *view ) {
    GtkNotebook        *view_book = (GtkNotebook *) view;
    GTaskViewPrivate   *private =  view->private;
    GtkTreeViewColumn  *column;
    GtkCellRenderer    *rend;
    GtkTreeView        *task_view;
    GtkScrolledWindow  *task_list_scroll;

    g_assert( private->task_view_list == NULL );

    task_list_scroll = GTK_SCROLLED_WINDOW( gtk_scrolled_window_new( NULL, NULL ) );

    gtk_scrolled_window_set_policy( task_list_scroll,
                                    GTK_POLICY_AUTOMATIC,
                                    GTK_POLICY_AUTOMATIC );

    task_view = GTK_TREE_VIEW( gtk_tree_view_new( ) );

    gtk_container_add( GTK_CONTAINER( task_list_scroll ),
                       GTK_WIDGET( task_view ) );

    gtk_notebook_append_page( view_book,
                              GTK_WIDGET( task_list_scroll ),
                              NULL );

    gtk_tree_view_set_headers_visible( task_view, TRUE );

    gtk_tree_view_set_rules_hint( task_view, TRUE );
    gtk_tree_view_set_reorderable( task_view, TRUE );

    rend = gtk_cell_renderer_text_new();

    gtk_tree_view_insert_column_with_attributes( task_view,
                                                 0,
                                                 "Title",
                                                 rend,
                                                 "text", COL_TITLE,
                                                 NULL );
    column = gtk_tree_view_get_column( task_view, 0 );
    gtk_tree_view_column_set_resizable( column, TRUE );

    gtk_tree_view_insert_column_with_attributes( task_view,
                                                 1,
                                                 "Status Message",
                                                 rend,
                                                 "text", COL_STATUS_MESSAGE,
                                                 NULL );
    column = gtk_tree_view_get_column( task_view, 1 );
    gtk_tree_view_column_set_resizable( column, TRUE );

    gtk_tree_view_insert_column_with_attributes( task_view,
                                                 2,
                                                 "Category",
                                                 rend,
                                                 "text", COL_CATEGORY,
                                                 NULL );

    column = gtk_tree_view_get_column( task_view, 2 );
    gtk_tree_view_column_set_resizable( column, TRUE );

    rend = GTK_CELL_RENDERER( gtask_cell_renderer_progress_new( ) );

    gtk_tree_view_insert_column_with_attributes( task_view,
                                                 3,
                                                 "Percent Completed",
                                                 rend,
                                                 "value", COL_PROGRESS,
                                                 "activity-state", COL_ACTIVITY_STATE,
                                                 NULL );

    column = gtk_tree_view_get_column( task_view, 3 );
    gtk_tree_view_column_set_resizable( column, TRUE );

    g_signal_connect( G_OBJECT( task_view ),
                      "button-press-event",
                      G_CALLBACK( gtask_view_button_press ),
                      view );

    private->task_view_list = task_view;
}

static void
test_close_not( gpointer data, GClosure *closure ) {
    g_debug( "===> in test_close_not for: %s", (gchar *) data );
    g_free( data );
}

static GtkWidget *
gtask_view_create_file_menu_subitem( GTaskView *view,
                                     GTaskFile *file,
                                     guint      bound )
{
    GTaskViewPrivate *private = view->private;
    GtkWidget        *file_item;
    const gchar      *tmp;
    gchar            *display_name;

    GSList           *preview_opts;
    GSList           *preview_opts_tmp;

    GdkPixbuf        *preview_pixbuf = NULL;

    const gchar      *file_uri = gtask_file_get_uri( file );

    /* determine the what to display in the label */
    tmp = gtask_file_get_description( file );

    if( tmp&& strlen( tmp ) )
        display_name = g_strdup( tmp );
    else
        display_name = g_path_get_basename( file_uri );

    file_item = gtk_image_menu_item_new_with_label( display_name );
    g_free( display_name );

    /* determine which thumbnail to load */
    preview_opts = gtask_ui_get_file_preview_options( file );

    preview_opts_tmp = preview_opts;

    while( preview_opts_tmp ) {
        GTaskPreview *preview = (GTaskPreview *) preview_opts_tmp->data;

        preview_pixbuf = gtask_preview_get_pixbuf( preview, 
                                                   private->thumbnailer );

        if( preview_pixbuf )
            break;

        preview_opts_tmp = g_slist_next( preview_opts_tmp );
    }

    gtask_preview_list_free( preview_opts );

    if( preview_pixbuf ) {
        GtkWidget *thumb_widget;
        GdkPixbuf *preview_scaled;

        preview_scaled = eel_gdk_pixbuf_scale_down_to_fit( preview_pixbuf,
                                                           bound,
                                                           bound );

        thumb_widget = gtk_image_new_from_pixbuf( preview_scaled );

        g_object_unref( (GObject *) preview_pixbuf );
        g_object_unref( (GObject *) preview_scaled );

        if( thumb_widget ) {
            gtk_image_menu_item_set_image( GTK_IMAGE_MENU_ITEM( file_item ),
                                           thumb_widget );
        }
    }

    if( gtask_file_get_completed( file ) ) {
        gtk_widget_set_sensitive( file_item, TRUE );

        g_signal_connect_data( G_OBJECT( file_item ),
                               "activate",
                               G_CALLBACK( gtask_view_open_file ),
                               g_strdup( file_uri ),
                               (GClosureNotify) test_close_not,
                               G_CONNECT_AFTER );
    } else {
        gtk_widget_set_sensitive( file_item, FALSE );
    }

    return file_item;
}

static GtkMenuItem *
gtask_view_create_file_menu_item( GTaskView         *view,
                                  GTaskFileListInfo *file_info,
                                  guint              bound )
{
    GtkMenuItem *item;
    GList       *files;

    g_assert( file_info != NULL );

    item = GTK_MENU_ITEM( gtk_image_menu_item_new_from_stock( GTK_STOCK_OPEN, NULL ) );

    files = file_info->files;

    if( file_info->num_files == 1 ) {
        GTaskFile   *file = (GTaskFile *) files->data;
        const gchar *file_uri = gtask_file_get_uri( file );

        if( gtask_file_get_completed( file ) ) {
            g_signal_connect_data( G_OBJECT( item ),
                                   "activate",
                                   G_CALLBACK( gtask_view_open_file ),
                                   g_strdup( file_uri ),
                                   (GClosureNotify) test_close_not,
                                   G_CONNECT_AFTER );

            gtk_widget_set_sensitive( (GtkWidget *) item, TRUE );
        } else {
            gtk_widget_set_sensitive( (GtkWidget *) item, FALSE );
        }
    } else if( file_info->num_files > 1 ) {
        GtkMenuShell *file_menu = GTK_MENU_SHELL( gtk_menu_new( ) );
            
        g_debug( "=====> %d %p", file_info->num_files, files );

        while( files ) {
            GTaskFile   *file = (GTaskFile *) files->data;
            GtkWidget   *file_sub;
            
            file_sub = gtask_view_create_file_menu_subitem( view, file, bound );

            gtk_menu_shell_append( file_menu, file_sub );
            files = g_list_next( files );
        }

        gtk_menu_item_set_submenu( item, (GtkWidget *) file_menu );
    } else {
        gtk_widget_set_sensitive( (GtkWidget *) item, FALSE );
    }

    return item;
}

static GtkMenuItem *
gtask_view_create_dir_menu_item( GTaskView         *view,
                                 GTaskFileListInfo *file_info )
{
    GtkMenuItem *item;
    GList       *dirs;

    g_assert( file_info != NULL );

    item = GTK_MENU_ITEM( gtk_menu_item_new_with_mnemonic( "Open Containing _Folder" ) );

    dirs = file_info->unique_directories;

    if( file_info->num_unique_directories == 1 ) {
        const gchar *dir_uri = (gchar *) dirs->data;

        g_signal_connect_data( G_OBJECT( item ),
                               "activate",
                               G_CALLBACK( gtask_view_open_dir ),
                               g_strdup( dir_uri ),
                               (GClosureNotify) test_close_not,
                               G_CONNECT_AFTER );

        gtk_widget_set_sensitive( (GtkWidget *) item, TRUE );
    } else if( file_info->num_unique_directories > 1 ) {
        GtkMenuShell *dir_menu = GTK_MENU_SHELL( gtk_menu_new( ) );
        GtkWidget    *dir_item;
            
        while( dirs ) {
            gchar *dir_uri = (gchar *) dirs->data;

            dir_item = gtk_menu_item_new_with_label( dir_uri );

            g_signal_connect_data( G_OBJECT( dir_item ),
                                   "activate",
                                   G_CALLBACK( gtask_view_open_dir ),
                                   g_strdup( dir_uri ),
                                   (GClosureNotify) test_close_not,
                                   G_CONNECT_AFTER );

            gtk_menu_shell_append( dir_menu, dir_item );
            dirs = g_list_next( dirs );
        }

        gtk_menu_item_set_submenu( item, (GtkWidget *) dir_menu );
    } else {
        gtk_widget_set_sensitive( (GtkWidget *) item, FALSE );
    }

    /** @todo link up the directory menu thumbnail (from nautilus) */

    return item;
}

static gboolean
gtask_view_button_press( GtkWidget *list,
                         GdkEventButton *event, 
                         gpointer data )
{
    GTaskView        *view = GTASK_VIEW( data );
    GTaskViewPrivate *private = view->private;

    g_debug( "IN: gtask_view_button_press" );

    /**
     * @todo eventually we need to check if the node that we clicked on
     *       was a group title or not
     */
    if( event->button == 3 && event->type == GDK_BUTTON_PRESS ) {
        GTaskGenericTask  *task;
        GTaskFileListInfo *file_info;

        GList             *files;
        GtkMenuShell      *task_menu;
        GtkTreePath       *path;
        GtkMenuItem       *item;
        GtkTreeIter        iter;
        GtkTreeModel      *task_model = GTK_TREE_MODEL( private->task_model );
        PangoLayout       *layout;
        guint              bound;

        /* figure out where the user clicked */
        if( !gtk_tree_view_get_path_at_pos( GTK_TREE_VIEW( list ),
                                            event->x,
                                            event->y,
                                            &path,
                                            NULL, NULL, NULL)) {
            return FALSE;
        }

        gtk_tree_model_get_iter( task_model, &iter, path );

        gtk_tree_model_get( task_model, &iter,
                            COL_TASK, &task,
                            -1 );

        files = gtask_generic_task_get_files( task );

        /** @todo render a piece of text in pango to determine
         *        the size that we need to scale our icons to
         */

        file_info = gtask_ui_get_file_list_info( files );

        task_menu = GTK_MENU_SHELL( gtk_menu_new( ) );

        g_signal_connect( task_menu,
                          "selection-done",
                          G_CALLBACK( gtk_object_destroy ),
                          NULL );

        layout = gtk_widget_create_pango_layout( list, NULL );

        pango_layout_get_pixel_size( layout, NULL, &bound );

        g_object_unref( (GObject *) layout );

        item = gtask_view_create_file_menu_item( view, file_info, bound );
        gtk_menu_shell_append( task_menu, (GtkWidget *) item );

        item = gtask_view_create_dir_menu_item( view, file_info );
        gtk_menu_shell_append( task_menu, (GtkWidget *) item );

        gtk_widget_show_all( (GtkWidget *) task_menu );

        gtk_menu_popup( (GtkMenu *) task_menu, NULL, NULL, NULL, NULL, 3, event->time);


        g_debug( "LEAVING: gtask_view_button_press" );

        return TRUE;
    }

    g_debug( "LEAVING: gtask_view_button_press" );

    return FALSE;
}

static void
close_error_window( GtkDialog *dialog, gint arg1, gpointer data ) {
    gtk_widget_destroy( GTK_WIDGET( dialog ) );
}

static void
gtask_view_open_file( GObject *obj, gpointer *data ) {
    char   *argv[3];
    char   *file = (char *) data;
    GError *error = NULL;

    argv[0] = "gnome-open";
    argv[1] = file;
    argv[2] = NULL;

    g_debug( "IN: gtask_view_open_file" );

    if( !g_spawn_async( NULL, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, NULL, &error ) )
    {
        GtkWidget *mess_win;

        /** @todo what if error is still NULL? */
        /** @todo HIG this puppy up */
        mess_win = gtk_message_dialog_new( NULL, 0,
                                GTK_MESSAGE_ERROR,
                                GTK_BUTTONS_OK,
                                error->message );

        g_signal_connect( mess_win,
                         "response",
                         G_CALLBACK( close_error_window ),
                         NULL );

        gtk_widget_show( mess_win );
    }

    g_debug( "LEAVING: gtask_view_open_file" );
}

/*
static void
gtask_view_open_file_properties( GObject *obj, gpointer *data ) {
    g_debug( "IN: gtask_view_open_file_properties" );
    g_debug( "LEAVING: gtask_view_open_file_properties" );
}
*/

static void
gtask_view_open_dir( GObject *obj, gpointer *data ) {
    gchar  *argv[3];
    gchar  *dir = (gchar *) data;
    GError *error = NULL;

    argv[0] = "nautilus";
    argv[1] = dir;
    argv[2] = NULL;

    g_debug( "IN: gtask_view_open_dir" );

    if( !g_spawn_async( NULL, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, NULL, &error ) )
    {
        GtkWidget *mess_win;

        /** @todo what if error is still NULL? */
        /** @todo HIG this puppy up */
        mess_win = gtk_message_dialog_new( NULL, 0,
                                GTK_MESSAGE_ERROR,
                                GTK_BUTTONS_OK,
                                error->message );

        g_signal_connect( mess_win,
                         "response",
                         G_CALLBACK( close_error_window ),
                         NULL );

        gtk_widget_show( mess_win );
    }

    g_debug( "LEAVING: gtask_view_open_dir" );
}

static void
gtask_view_set_property( GObject *object,
                         guint param_id,
                         const GValue *value,
                         GParamSpec *pspec )
{
    GTaskView *view = GTASK_VIEW( object );

    switch( param_id ) {
        case PROP_VIEW_MODE:
            gtask_view_set_view_mode( view, g_value_get_enum( value ) );
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID( object, param_id, pspec );
    }
}

static void
gtask_view_get_property( GObject *object,
                         guint param_id,
                         GValue *value,
                         GParamSpec *pspec )
{
    GTaskView *view = GTASK_VIEW( object );

    switch( param_id ) {
        case PROP_VIEW_MODE:
            g_value_set_enum( value, gtask_view_get_view_mode( view ) );
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID( object, param_id, pspec );
    }
}

GTaskView *
gtask_view_new( ) {
    return GTASK_VIEW( g_object_new( GTASK_VIEW_TYPE, NULL ) );
}

void
gtask_view_set_view_mode( GTaskView *view, GTaskViewMode mode ) {
    GTaskViewPrivate *private;

    g_return_if_fail( GTASK_IS_VIEW( view ) );

    g_debug( "IN: gtask_view_set_view_mode" );

    private = view->private;

    if( mode == private->view_mode ) {
        g_debug( "LEAVING: gtask_view_set_view_mode - new mode same as old one" );
        return;
    }

    private->view_mode = mode;

    if( mode == GTASK_VIEW_MODE_BASIC ) {
        gtk_tree_view_set_model( private->task_view_basic,
                                 (GtkTreeModel *) private->task_model_filter );

        gtk_notebook_set_current_page( (GtkNotebook *) view, 0 );

        gtk_tree_view_set_model( private->task_view_list, NULL );
    } else {
        gtk_tree_view_set_model( private->task_view_list,
                                 (GtkTreeModel *) private->task_model_filter );

        gtk_notebook_set_current_page( (GtkNotebook *) view, 1 );

        gtk_tree_view_set_model( private->task_view_basic, NULL );
    }

    g_object_notify( (GObject *) view, "view-mode" );

    g_debug( "LEAVING: gtask_view_set_view_mode - mode changed" );
}

GTaskViewMode
gtask_view_get_view_mode( GTaskView *view ) {
    g_return_val_if_fail( GTASK_IS_VIEW( view ), GTASK_VIEW_MODE_BASIC );

    return view->private->view_mode;
}

void
gtask_view_add_task( GTaskView *view, GTaskGenericTask *task ) {
    GTaskViewPrivate *private = view->private;
    const char       *task_id;
    GTaskTaskIter    *task_iter;
    guint             signal_id;
    GList            *task_list = NULL;

    g_return_if_fail( GTASK_IS_VIEW( view ) );
    g_return_if_fail( GTASK_IS_GENERIC_TASK( task ) );

    g_warning( "IN: gtask_view_add_task" );

    task_id = gtask_generic_task_get_id( task );

    task_iter = (GTaskTaskIter *) g_hash_table_lookup( private->task_hash,
                                                       task_id );

    /** @todo store row references not iters! */

    /* not in the model, insert it in */
    if( !task_iter ) {
        task_iter = g_new0( GTaskTaskIter, 1 );

        g_object_ref( task );
        task_iter->task = task;
        
        gtk_tree_store_append( private->task_model, &(task_iter->iter), NULL );

        /* set a default activity state */
        gtk_tree_store_set( private->task_model, &(task_iter->iter),
                            COL_ACTIVITY_STATE, GTASK_ACTIVITY_STATE_ACTIVE, 
                            -1 );

        g_hash_table_insert( private->task_hash,
                             g_strdup( task_id ),
                             task_iter );

        signal_id = GTASK_VIEW_GET_CLASS( view )->tasks_added_signal_id;
    } else {
        /** @todo right now tasks send all their data across the line each time
         *         this may change in the future
         */
        g_object_unref( task_iter->task );
        g_object_ref( task );
        task_iter->task = task;

        signal_id = GTASK_VIEW_GET_CLASS( view )->tasks_updated_signal_id;
    }

    gtask_view_update_task( view, task_iter );

    task_list = g_list_append( task_list, task );

    g_signal_emit( view, signal_id, 0, task_list );

    g_list_free( task_list );

    g_warning( "LEAVING: gtask_view_add_task" );
}

static void
gtask_view_update_task_thumbnail( GTaskView *view, GTaskTaskIter *task_iter ) {
    GTaskViewPrivate   *private = view->private;
    GtkTreeIter        iter = task_iter->iter;
    GTaskGenericTask   *task = task_iter->task;
    GtkTreeModel       *model = GTK_TREE_MODEL( view->private->task_model );

    GTaskActivityState  state;
    const char         *old_thumb_uri;
    GValue              value = { 0, };

    GSList              *preview_list;
    GSList              *preview_list_tmp;

    g_debug( "IN: gtask_view_update_task_thumbnail" );

    state = gtask_generic_task_get_activity_state( task );

    gtk_tree_model_get_value( model, &iter, COL_OLD_THUMB_URI, &value );
    old_thumb_uri = g_value_get_string( &value );

    if( !old_thumb_uri )
        old_thumb_uri = "";

    preview_list = gtask_ui_get_task_preview_options( task );

    preview_list_tmp = preview_list;

    while( preview_list_tmp ) {
        GTaskPreview *preview = (GTaskPreview *) preview_list_tmp->data;
        GdkPixbuf    *preview_pixbuf;

        if( strcmp( preview->uri, old_thumb_uri ) == 0 )
            break;

                
        preview_pixbuf = gtask_preview_get_pixbuf( preview,
                                                   private->thumbnailer );

        if( preview_pixbuf ) {
            gtk_tree_store_set( GTK_TREE_STORE( private->task_model ), &iter,
                                COL_THUMBNAIL, preview_pixbuf,
                                COL_OLD_THUMB_URI, preview->uri,
                                COL_IS_THUMBNAIL, preview->is_thumbnail,
                                -1 );

            g_object_unref( (GObject *) preview_pixbuf );
        }

        preview_list_tmp = g_slist_next( preview_list_tmp );
    }

    gtask_preview_list_free( preview_list );

    g_value_unset( &value );

    g_debug( "LEAVING: gtask_view_update_task_thumbnail" );
}

static void
gtask_view_update_task( GTaskView *view, GTaskTaskIter *task_iter ) {
    GtkTreeIter         iter = task_iter->iter;
    GTaskGenericTask   *task = task_iter->task;

    GTaskViewPrivate   *private = view->private;
    GTaskActivityState  prev_activity_state;
    GTaskActivityState  curr_activity_state;

    const char         *status_msg;
    const char         *error_msg;
    const char         *title;
    const char         *category;
    const char         *thumb_uri;
    gfloat              percent;
    glong               time_left;

    GValue              val = { 0, };

    g_debug( "IN: gtask_view_update_task" );

    gtk_tree_model_get_value( GTK_TREE_MODEL( private->task_model ), &iter,
                              COL_ACTIVITY_STATE, &val );

    prev_activity_state = g_value_get_enum( &val );

    g_value_unset( &val );

    curr_activity_state = gtask_generic_task_get_activity_state( task );

    g_warning( "==========> %s : prev-%d : curr-%d", gtask_generic_task_get_title( task ), prev_activity_state, curr_activity_state );

    if( curr_activity_state == GTASK_ACTIVITY_STATE_COMPLETED &&
        prev_activity_state == GTASK_ACTIVITY_STATE_COMPLETED )
    {
        gtask_view_update_task_thumbnail( view, task_iter );
        /** @todo as of right now no more updates are done on completed
         *        tasks. Perhaps in the future we should allow for minor
         *        updates - ie, status, etc
         */
        g_debug( "LEAVING: gtask_view_update_task - prev status was COMPLETED" );

        return;
    }

    status_msg = gtask_generic_task_get_status_message( task );
    error_msg = gtask_generic_task_get_error_message( task );

    if( curr_activity_state == GTASK_ACTIVITY_STATE_INACTIVE &&
        prev_activity_state == GTASK_ACTIVITY_STATE_INACTIVE )
    {
        gtk_tree_store_set( private->task_model, &iter,
                            COL_TASK, task,
                            COL_STATUS_MESSAGE, status_msg,
                            COL_ERROR_MESSAGE, error_msg,
                            -1 );
        g_debug( "LEAVING: gtask_view_update_task - last 2 status were INACTIVE" );

        return;
    }

    gtask_view_update_task_thumbnail( view, task_iter );

    percent = gtask_generic_task_get_percent_done( task );;
    title = gtask_generic_task_get_title( task );
    category = gtask_generic_task_get_category( task );
    thumb_uri = gtask_generic_task_get_thumbnail_uri( task );
    time_left = gtask_generic_task_get_time_left( task );

    gtk_tree_store_set( private->task_model, &iter,
                        COL_TASK, task,
                        COL_PROGRESS, percent,
                        COL_TITLE, title,
                        COL_STATUS_MESSAGE, status_msg,
                        COL_CATEGORY, category,
                        COL_ACTIVITY_STATE, curr_activity_state,
                        COL_TIME_LEFT, time_left,
                        -1 );

    g_debug( "LEAVING: gtask_view_update_task - task updated" );
}

static gboolean
remove_task( gpointer key, gpointer value, gpointer user_data ) {
    struct GTaskGetTaskFoo *boo = (struct GTaskGetTaskFoo *) user_data;

    GTaskView          *view = GTASK_VIEW( boo->view );
    GTaskTaskIter      *iter = (GTaskTaskIter *) value;
    GTaskGenericTask   *task = iter->task;
    GTaskActivityState  state = gtask_generic_task_get_activity_state( task );
    gboolean            remove = FALSE;

    switch( state ) {
        case GTASK_ACTIVITY_STATE_ACTIVE:
            if( boo->with_active )
                remove = TRUE;
            break;
        case GTASK_ACTIVITY_STATE_INACTIVE:
            if( boo->with_inactive )
                remove = TRUE;
            break;
        case GTASK_ACTIVITY_STATE_COMPLETED:
            if( boo->with_completed )
                remove = TRUE;
            break;
        default:
            break;
    }

    if( remove ) {
        gtk_tree_store_remove( view->private->task_model, &(iter->iter) );

        g_object_ref( task );

        boo->tasks = g_list_append( boo->tasks, task );

        return TRUE;
    }

    return FALSE;
}

void
gtask_view_clear_tasks( GTaskView *view, GTaskActivityState with_state ) {
    GTaskViewPrivate *private;
    struct GTaskGetTaskFoo boo = { 0, NULL, NULL };

    g_debug( "IN: gtask_view_clear_tasks" );

    boo.view = view;
    boo.with_active = ( with_state & GTASK_ACTIVITY_STATE_ACTIVE ) != 0;
    boo.with_inactive = ( with_state & GTASK_ACTIVITY_STATE_INACTIVE ) != 0;
    boo.with_completed = ( with_state & GTASK_ACTIVITY_STATE_COMPLETED ) != 0;

    g_return_if_fail( GTASK_IS_VIEW( view ) );

    private = view->private;

    /* clear the model, and the lookup hash
     * clearing the model is a side effect in remove_task() */
    g_hash_table_foreach_remove( private->task_hash, remove_task, &boo );

    g_signal_emit( view,
                   GTASK_VIEW_GET_CLASS( view )->tasks_removed_signal_id,
                   0,
                   boo.tasks );

    g_list_foreach( boo.tasks, (GFunc) g_object_unref, NULL );

    g_debug( "LEAVING: gtask_view_clear_tasks" );
}

void
gtask_view_set_filter_string( GTaskView *view, const gchar *str ) {
    GTaskViewPrivate *private;
    PangoLogAttr     *attrs;
    char             *tmp_string;
    char             *norm_string;
    guint             word_start = 0;
    guint             str_len;
    guint             i;

    g_return_if_fail( GTASK_IS_VIEW( view ) );
    g_return_if_fail( str != NULL );
    g_return_if_fail( g_utf8_validate( str, -1, NULL ) );

    g_debug( "IN: gtask_view_set_filter_string" );

    private = view->private;

    tmp_string = g_utf8_strdown( str, -1 );
    norm_string = g_utf8_normalize( tmp_string, -1, G_NORMALIZE_ALL );

    for( i = 0; i < private->filter_seed->len; i++ ) {
        g_debug( "ffff" );
        g_free( private->filter_seed->pdata[i] );
    }

    g_ptr_array_set_size( private->filter_seed, 0 );

    str_len = g_utf8_strlen( norm_string, -1 );
    attrs = g_new0( PangoLogAttr, str_len + 1 );

    /** @todo get the current PangoLanguage? */
    pango_get_log_attrs( norm_string, -1, -1, NULL, attrs, str_len + 1 );

    g_debug( "start splitting string" );
    for( i = 0; i < str_len + 1; i++ ) {
        char *start_word;
        char *end_word;

        if( attrs[i].is_word_end ) {
            char *word;

            start_word = g_utf8_offset_to_pointer( norm_string, word_start );
            end_word = g_utf8_offset_to_pointer( norm_string, i );

            word  = g_strndup( start_word, end_word - start_word );

            g_ptr_array_add( private->filter_seed, word );
        }

        if( attrs[i].is_word_start )
            word_start = i;
    }
    g_debug( "end splitting string" );

    g_string_assign( private->filter_string, str );

    g_free( attrs );
    g_free( tmp_string );
    g_free( norm_string );

//    g_object_notify( (GObject *) view, "filter-string" );

    gtk_tree_model_filter_refilter( private->task_model_filter );

    g_debug( "LEAVING: gtask_view_set_filter_string" );
}

const gchar *
gtask_view_get_filter_string( GTaskView *view ) {
    g_return_val_if_fail( GTASK_IS_VIEW( view ), "" );

    return view->private->filter_string->str;
}

static void
sum_tasks( gpointer key, gpointer value, gpointer user_data ) {
    struct GTaskGetTaskFoo  *boo = (struct GTaskGetTaskFoo *) user_data;

    GTaskTaskIter *iter = (GTaskTaskIter *) value;

    GTaskGenericTask *task = GTASK_GENERIC_TASK( iter->task );
    GTaskActivityState state = gtask_generic_task_get_activity_state( task );

    switch( state ) {
        case GTASK_ACTIVITY_STATE_ACTIVE:
            if( boo->with_active )
                boo->num++;
            break;
        case GTASK_ACTIVITY_STATE_INACTIVE:
            if( boo->with_inactive )
                boo->num++;
            break;
        case GTASK_ACTIVITY_STATE_COMPLETED:
            if( boo->with_completed )
                boo->num++;
            break;
        default:
            break;
    }
}

guint
gtask_view_get_number_of_tasks( GTaskView *view,
                                GTaskActivityState with_state )
{
    struct GTaskGetTaskFoo boo = { 0, NULL, NULL, };

    g_debug( "IN: gtask_view_get_number_of_tasks" );

    boo.with_active = ( with_state & GTASK_ACTIVITY_STATE_ACTIVE ) != 0;
    boo.with_inactive = ( with_state & GTASK_ACTIVITY_STATE_INACTIVE ) != 0;
    boo.with_completed = ( with_state & GTASK_ACTIVITY_STATE_COMPLETED ) != 0;

    g_return_val_if_fail( GTASK_IS_VIEW( view ), 0 );

    g_hash_table_foreach( view->private->task_hash, sum_tasks, &boo );

    g_debug( "LEAVING: gtask_view_get_number_of_tasks" );

    return boo.num;
}

/* @} */
